import _defineProperty from '@babel/runtime/helpers/defineProperty';
import _objectWithoutProperties from '@babel/runtime/helpers/objectWithoutProperties';

/**
 * Checks if `value` is a `function`.
 * @param value The value to check
 * @returns Returns `true` if `value` is a function, else `false`.
 */
function isFunction(value) {
  return typeof value === "function";
}
var slice = Array.prototype.slice;
var noop = () => {};
/** set dom attribute for consistent behavior */
var updateDomAttr = (dom, name, value) => {
  var isAria = false;
  if (value != null && (value !== false || (isAria = name.indexOf('aria-') === 0))) {
    dom.setAttribute(name, !isAria && typeof value === 'boolean'
    // true but not aria-*
    ? ''
    // falsy aria-* values
    : value);
  } else {
    dom.removeAttribute(name);
  }
};

/**
 * Find the closest error boundary to a thrown error and call it
 * @param {object} error The thrown value
 * @param vnode The vnode that threw
 * the error that was caught (except for unmounting when this parameter
 * is the highest parent that was being unmounted)
 * @param oldVNode
 * @param errorInfo
 */
function _catchError(error, vnode, oldVNode, errorInfo) {
  var component, ctor, handled;
  for (; vnode = vnode._parent;) {
    if ((component = vnode._component) && !component._processingException) {
      try {
        ctor = component.constructor;
        if (ctor && ctor.getDerivedStateFromError != null) {
          Object.assign(component, ctor.getDerivedStateFromError(error));
          handled = component._dirty;
        }
        if (component.componentDidCatch != null) {
          component.componentDidCatch(error, errorInfo || {});
          handled = component._dirty;
        }

        // This is an error boundary. Mark it as having bailed out, and whether it was mid-hydration.
        if (handled) {
          return component._pendingError = component;
        }
      } catch (e) {
        error = e;
      }
    }
  }
  throw error;
}

/**
 * The `option` object can potentially contain callback functions
 * that are called during various stages of our renderer. This is the
 * foundation on which all our addons like `quark/debug`, `quark/compat`,
 * and `quark/hooks` are based on. See the `Options` type in `internal.d.ts`
 * for a full list of available option hooks (most editors/IDEs allow you to
 * ctrl+click or cmd+click on mac the type definition below).
 */
var options = {
  _catchError
};

var vnodeId = 0;

/**
 * Create an virtual node (used for JSX)
 * @param type The node name or Component
 * constructor for this virtual node
 * @param {object | null | undefined} props The properties of the virtual node
 * @param children The children of the virtual node
 * @returns
 */
function createElement(type, props, children) {
  var normalizedProps = {},
    key,
    ref,
    i;
  for (i in props) {
    if (i == 'key') key = props[i];else if (i == 'ref') ref = props[i];else normalizedProps[i] = props[i];
  }
  if (arguments.length > 2) {
    normalizedProps.children = arguments.length > 3 ? slice.call(arguments, 2) : children;
  }

  // If a Component VNode, check for and apply defaultProps
  // Note: type may be undefined in development, must never error here.
  if (typeof type == 'function' && type.defaultProps != null) {
    for (i in type.defaultProps) {
      if (normalizedProps[i] === undefined) {
        normalizedProps[i] = type.defaultProps[i];
      }
    }
  }
  return createVNode(type, normalizedProps, key, ref, null);
}

/**
 * Create a VNode (used internally by quark)
 * @param type The node name or Component
 * Constructor for this virtual node
 * @param {object | string | number | null} props The properties of this virtual node.
 * If this virtual node represents a text node, this is the text of the node (string or number).
 * @param {string | number | null} key The key for this virtual node, used when
 * diffing it against its children
 * @param ref The ref property that will
 * receive a reference to its created child
 * @returns
 */
function createVNode(type, props, key, ref, original) {
  // V8 seems to be better at detecting type shapes if the object is allocated from the same call site
  // Do not inline into createElement and coerceToVNode!
  var vnode = {
    type,
    props,
    key,
    ref,
    _children: null,
    _parent: null,
    _depth: 0,
    _dom: null,
    // _nextDom must be initialized to undefined b/c it will eventually
    // be set to dom.nextSibling which can return `null` and it is important
    // to be able to distinguish between an uninitialized _nextDom and
    // a _nextDom that has been set to `null`
    _nextDom: undefined,
    _component: null,
    _hydrating: null,
    constructor: undefined,
    _original: original == null ? ++vnodeId : original
  };

  // Only invoke the vnode hook if this was *not* a direct copy:
  if (original == null && options.vnode != null) options.vnode(vnode);
  return vnode;
}
function Fragment(props) {
  return props.children;
}

var EMPTY_OBJ = {};
var EMPTY_ARR = [];
var IS_NON_DIMENSIONAL = /acit|ex(?:s|g|n|p|$)|rph|grid|ows|mnc|ntw|ine[ch]|zoo|^ord|itera/i;

/**
 * Base Component class.
 * @param {object} props The initial component props
 */
function Component(props) {
  this.props = props;
  this._dirty = false;
}

/**
 * Accepts `props`, and returns a new Virtual DOM tree to build.
 * Virtual DOM is generally constructed via [JSX](http://jasonformat.com/wtf-is-jsx).
 * @param {object} props Props (eg: JSX attributes) received from parent
 * element/component
 * @returns {import('./index').ComponentChildren | void}
 */
Component.prototype.render = Fragment;

/**
 * @param vnode
 * @param {number | null} [childIndex]
 */
function getDomSibling(vnode, childIndex) {
  if (childIndex == null) {
    // Use childIndex==null as a signal to resume the search from the vnode's sibling
    return vnode._parent ? getDomSibling(vnode._parent, vnode._parent._children.indexOf(vnode) + 1) : null;
  }
  var sibling;
  for (; childIndex < vnode._children.length; childIndex++) {
    sibling = vnode._children[childIndex];
    if (sibling != null && sibling._dom != null) {
      // Since updateParentDomPointers keeps _dom pointer correct,
      // we can rely on _dom to tell us if this subtree contains a
      // rendered DOM node, and what the first rendered DOM node is
      return sibling._dom;
    }
  }

  // If we get here, we have not found a DOM node in this vnode's children.
  // We must resume from this vnode's sibling (in it's parent _children array)
  // Only climb up and search the parent if we aren't searching through a DOM
  // VNode (meaning we reached the DOM parent of the original vnode that began
  // the search)
  return typeof vnode.type == 'function' ? getDomSibling(vnode) : null;
}

/**
 * Diff the children of a virtual node
 * @param parentDom The DOM element whose
 * children are being diffed
 * @param renderResult
 * @param newParentVNode The new virtual
 * node whose children should be diff'ed against oldParentVNode
 * @param oldParentVNode The old virtual
 * node whose children should be diff'ed against newParentVNode
 * @param {boolean} isSvg Whether or not this DOM node is an SVG node
 * @param excessDomChildren
 * @param oldDom The current attached DOM
 * element any new dom elements should be placed around. Likely `null` on first
 * render (except when hydrating). Can be a sibling DOM element when diffing
 * Fragments that have siblings. In most cases, it starts out as `oldChildren[0]._dom`.
 */
function diffChildren(parentDom, renderResult, newParentVNode, oldParentVNode, isSvg, excessDomChildren, oldDom) {
  var i, j, oldVNode, childVNode, newDom, firstChildDom, refs;

  // This is a compression of oldParentVNode!=null && oldParentVNode != EMPTY_OBJ && oldParentVNode._children || EMPTY_ARR
  // as EMPTY_OBJ._children should be `undefined`.
  var oldChildren = oldParentVNode && oldParentVNode._children || EMPTY_ARR;
  var oldChildrenLength = oldChildren.length;
  newParentVNode._children = [];
  for (i = 0; i < renderResult.length; i++) {
    childVNode = renderResult[i];
    if (childVNode == null || typeof childVNode == 'boolean' || typeof childVNode == 'function') {
      childVNode = newParentVNode._children[i] = null;
    }

    // If this newVNode is being reused (e.g. <div>{reuse}{reuse}</div>) in the same diff,
    // or we are rendering a component copy the oldVNodes so it can have
    // it's own DOM & etc. pointers
    else if (typeof childVNode == 'string' || typeof childVNode == 'number' ||
    // eslint-disable-next-line valid-typeof
    typeof childVNode == 'bigint') {
      childVNode = newParentVNode._children[i] = createVNode(null, childVNode, null, null, childVNode);
    } else if (Array.isArray(childVNode)) {
      childVNode = newParentVNode._children[i] = createVNode(Fragment, {
        children: childVNode
      }, null, null, null);
    } else if (childVNode._depth > 0) {
      // VNode is already in use, clone it. This can happen in the following
      // scenario:
      //   const reuse = <div />
      //   <div>{reuse}<span />{reuse}</div>
      childVNode = newParentVNode._children[i] = createVNode(childVNode.type, childVNode.props, childVNode.key, childVNode.ref ? childVNode.ref : null, childVNode._original);
    } else {
      childVNode = newParentVNode._children[i] = childVNode;
    }

    // Terser removes the `continue` here and wraps the loop body
    // in a `if (childVNode) { ... } condition
    if (childVNode == null) {
      continue;
    }
    childVNode._parent = newParentVNode;
    childVNode._depth = newParentVNode._depth + 1;

    // Check if we find a corresponding element in oldChildren.
    // If found, delete the array item by setting to `undefined`.
    // We use `undefined`, as `null` is reserved for empty placeholders
    // (holes).
    oldVNode = oldChildren[i];
    if (oldVNode === null || oldVNode && childVNode.key == oldVNode.key && childVNode.type === oldVNode.type) {
      oldChildren[i] = undefined;
    } else {
      // Either oldVNode === undefined or oldChildrenLength > 0,
      // so after this loop oldVNode == null or oldVNode is a valid value.
      for (j = 0; j < oldChildrenLength; j++) {
        oldVNode = oldChildren[j];
        // If childVNode is unkeyed, we only match similarly unkeyed nodes, otherwise we match by key.
        // We always match by type (in either case).
        if (oldVNode && childVNode.key == oldVNode.key && childVNode.type === oldVNode.type) {
          oldChildren[j] = undefined;
          break;
        }
        oldVNode = null;
      }
    }
    oldVNode = oldVNode || EMPTY_OBJ;

    // Morph the old element into the new one, but don't append it to the dom yet
    diff(parentDom, childVNode, oldVNode, isSvg, excessDomChildren, oldDom);
    newDom = childVNode._dom;
    if ((j = childVNode.ref) && oldVNode.ref != j) {
      if (!refs) refs = [];
      if (oldVNode.ref) refs.push(oldVNode.ref, null, childVNode);
      refs.push(j, childVNode._component || newDom, childVNode);
    }
    if (newDom != null) {
      if (firstChildDom == null) {
        firstChildDom = newDom;
      }
      if (typeof childVNode.type == 'function' && childVNode._children === oldVNode._children) {
        childVNode._nextDom = oldDom = reorderChildren(childVNode, oldDom, parentDom);
      } else {
        oldDom = placeChild(parentDom, childVNode, oldVNode, oldChildren, newDom, oldDom);
      }
      if (typeof newParentVNode.type == 'function') {
        // Because the newParentVNode is Fragment-like, we need to set it's
        // _nextDom property to the nextSibling of its last child DOM node.
        //
        // `oldDom` contains the correct value here because if the last child
        // is a Fragment-like, then oldDom has already been set to that child's _nextDom.
        // If the last child is a DOM VNode, then oldDom will be set to that DOM
        // node's nextSibling.
        newParentVNode._nextDom = oldDom;
      }
    } else if (oldDom && oldVNode._dom == oldDom && oldDom.parentNode != parentDom) {
      // The above condition is to handle null placeholders. See test in placeholder.test.js:
      // `efficiently replace null placeholders in parent rerenders`
      oldDom = getDomSibling(oldVNode);
    }
  }
  newParentVNode._dom = firstChildDom;

  // Remove remaining oldChildren if there are any.
  for (i = oldChildrenLength; i--;) {
    if (oldChildren[i] != null) {
      if (typeof newParentVNode.type == 'function' && oldChildren[i]._dom != null && oldChildren[i]._dom == newParentVNode._nextDom) {
        // If the newParentVNode.__nextDom points to a dom node that is about to
        // be unmounted, then get the next sibling of that vnode and set
        // _nextDom to it
        newParentVNode._nextDom = getLastDom(oldParentVNode).nextSibling;
      }
      unmount(oldChildren[i], oldChildren[i]);
    }
  }

  // Set refs only after unmount
  if (refs) {
    for (i = 0; i < refs.length; i++) {
      applyRef(refs[i], refs[++i], refs[++i]);
    }
  }
}
function reorderChildren(childVNode, oldDom, parentDom) {
  // Note: VNodes in nested suspended trees may be missing _children.
  var c = childVNode._children;
  var tmp = 0;
  for (; c && tmp < c.length; tmp++) {
    var vnode = c[tmp];
    if (vnode) {
      // We typically enter this code path on sCU bailout, where we copy
      // oldVNode._children to newVNode._children. If that is the case, we need
      // to update the old children's _parent pointer to point to the newVNode
      // (childVNode here).
      vnode._parent = childVNode;
      if (typeof vnode.type == 'function') {
        oldDom = reorderChildren(vnode, oldDom, parentDom);
      } else {
        oldDom = placeChild(parentDom, vnode, vnode, c, vnode._dom, oldDom);
      }
    }
  }
  return oldDom;
}
function placeChild(parentDom, childVNode, oldVNode, oldChildren, newDom, oldDom) {
  var nextDom;
  if (childVNode._nextDom !== undefined) {
    // Only Fragments or components that return Fragment like VNodes will
    // have a non-undefined _nextDom. Continue the diff from the sibling
    // of last DOM child of this child VNode
    nextDom = childVNode._nextDom;

    // Eagerly cleanup _nextDom. We don't need to persist the value because
    // it is only used by `diffChildren` to determine where to resume the diff after
    // diffing Components and Fragments. Once we store it the nextDOM local var, we
    // can clean up the property
    childVNode._nextDom = undefined;
  } else if (oldVNode == null || newDom != oldDom || newDom.parentNode == null) {
    outer: if (oldDom == null || oldDom.parentNode !== parentDom) {
      parentDom.appendChild(newDom);
      nextDom = null;
    } else {
      // `j<oldChildrenLength; j+=2` is an alternative to `j++<oldChildrenLength/2`
      for (var sibDom = oldDom, j = 0; (sibDom = sibDom.nextSibling) && j < oldChildren.length; j += 1) {
        if (sibDom == newDom) {
          break outer;
        }
      }
      parentDom.insertBefore(newDom, oldDom);
      nextDom = oldDom;
    }
  }

  // If we have pre-calculated the nextDOM node, use it. Else calculate it now
  // Strictly check for `undefined` here cuz `null` is a valid value of `nextDom`.
  // See more detail in create-element.js:createVNode
  if (nextDom !== undefined) {
    oldDom = nextDom;
  } else {
    oldDom = newDom.nextSibling;
  }
  return oldDom;
}

/**
 * @param vnode
 */
function getLastDom(vnode) {
  if (vnode.type == null || typeof vnode.type === 'string') {
    return vnode._dom;
  }
  if (vnode._children) {
    for (var i = vnode._children.length - 1; i >= 0; i--) {
      var child = vnode._children[i];
      if (child) {
        var lastDom = getLastDom(child);
        if (lastDom) {
          return lastDom;
        }
      }
    }
  }
  return null;
}

/**
 * Diff the old and new properties of a VNode and apply changes to the DOM node
 * @param dom The DOM node to apply
 * changes to
 * @param {object} newProps The new props
 * @param {object} oldProps The old props
 * @param {boolean} isSvg Whether or not this node is an SVG node
 */
function diffProps(dom, newProps, oldProps, isSvg) {
  var i;
  for (i in oldProps) {
    if (i !== 'children' && i !== 'key' && !(i in newProps)) {
      setProperty(dom, i, null, oldProps[i], isSvg);
    }
  }
  for (i in newProps) {
    if (i !== 'children' && i !== 'key' && i !== 'value' && i !== 'checked' && oldProps[i] !== newProps[i]) {
      setProperty(dom, i, newProps[i], oldProps[i], isSvg);
    }
  }
}
function setStyle(style, key, value) {
  if (key[0] === '-') {
    style.setProperty(key, value == null ? '' : value);
  } else if (value == null) {
    style[key] = '';
  } else if (typeof value != 'number' || IS_NON_DIMENSIONAL.test(key)) {
    style[key] = value;
  } else {
    style[key] = value + 'px';
  }
}

/**
 * Set a property value on a DOM node
 * @param dom The DOM node to modify
 * @param {string} name The name of the property to set
 * @param value The value to set the property to
 * @param oldValue The old value the property had
 * @param {boolean} isSvg Whether or not this DOM node is an SVG node or not
 */
function setProperty(dom, name, value, oldValue, isSvg) {
  var useCapture;
  o: if (name === 'style') {
    if (typeof value == 'string') {
      dom.style.cssText = value;
    } else {
      if (typeof oldValue == 'string') {
        dom.style.cssText = oldValue = '';
      }
      if (oldValue) {
        for (name in oldValue) {
          if (!(value && name in value)) {
            setStyle(dom.style, name, '');
          }
        }
      }
      if (value) {
        for (name in value) {
          if (!oldValue || value[name] !== oldValue[name]) {
            setStyle(dom.style, name, value[name]);
          }
        }
      }
    }
  }
  // Benchmark for comparison: https://esbench.com/bench/574c954bdb965b9a00965ac6
  else if (name[0] === 'o' && name[1] === 'n') {
    useCapture = name !== (name = name.replace(/Capture$/, ''));

    // Infer correct casing for DOM built-in events:
    if (name.toLowerCase() in dom) name = name.toLowerCase().slice(2);else name = name.slice(2);
    if (!dom._listeners) dom._listeners = {};
    dom._listeners[name + useCapture] = value;
    if (value) {
      if (!oldValue) {
        var handler = useCapture ? eventProxyCapture : eventProxy;
        dom.addEventListener(name, handler, useCapture);
      }
    } else {
      var _handler = useCapture ? eventProxyCapture : eventProxy;
      dom.removeEventListener(name, _handler, useCapture);
    }
  } else if (name !== 'dangerouslySetInnerHTML') {
    if (isSvg) {
      // Normalize incorrect prop usage for SVG:
      // - xlink:href / xlinkHref --> href (xlink:href was removed from SVG and isn't needed)
      // - className --> class
      name = name.replace(/xlink(H|:h)/, 'h').replace(/sName$/, 's');
    } else if (name !== 'width' && name !== 'height' && name !== 'href' && name !== 'list' && name !== 'form' &&
    // Default value in browsers is `-1` and an empty string is
    // cast to `0` instead
    name !== 'tabIndex' && name !== 'download' && name in dom) {
      try {
        var _dom$_isInternalProp;
        // * update internal props here
        if ((_dom$_isInternalProp = dom._isInternalProp) !== null && _dom$_isInternalProp !== void 0 && _dom$_isInternalProp.call(dom, name)) {
          dom[name] = value;
        } else {
          dom[name] = value == null ? '' : value;
        }

        // labelled break is 1b smaller here than a return statement (sorry)
        break o;
      } catch (e) {}
    }

    // ARIA-attributes have a different notion of boolean values.
    // The value `false` is different from the attribute not
    // existing on the DOM, so we can't remove it. For non-boolean
    // ARIA-attributes we could treat false as a removal, but the
    // amount of exceptions would cost us too many bytes. On top of
    // that other VDOM frameworks also always stringify `false`.

    if (typeof value === 'function') ; else {
      updateDomAttr(dom, name, value);
    }
  }
}

/**
 * Proxy an event to hooked event handlers
 * @param {Event} e The event object from the browser
 * @private
 */
function eventProxy(e) {
  return this._listeners[e.type + false](options.event ? options.event(e) : e);
}
function eventProxyCapture(e) {
  return this._listeners[e.type + true](options.event ? options.event(e) : e);
}

/**
 * Diff two virtual nodes and apply proper changes to the DOM
 * @param parentDom The parent of the DOM element
 * @param newVNode The new virtual node
 * @param oldVNode The old virtual node
 * @param {boolean} isSvg Whether or not this element is an SVG node
 * @param excessDomChildren
 * @param oldDom The current attached DOM
 * element any new dom elements should be placed around. Likely `null` on first
 * render (except when hydrating). Can be a sibling DOM element when diffing
 * Fragments that have siblings. In most cases, it starts out as `oldChildren[0]._dom`.
 */
function diff(parentDom, newVNode, oldVNode, isSvg, excessDomChildren, oldDom) {
  var tmp,
    newType = newVNode.type;

  // When passing through createElement it assigns the object
  // constructor as undefined. This to prevent JSON-injection.
  if (newVNode.constructor !== undefined) return null;
  if (tmp = options._diff) tmp(newVNode);
  try {
    outer: if (typeof newType == 'function') {
      var c, isNew, oldProps, clearProcessingException;
      var newProps = newVNode.props;

      // Get component and set it to `c`
      if (oldVNode._component) {
        c = newVNode._component = oldVNode._component;
        clearProcessingException = c._processingException = c._pendingError;
      } else {
        if ('prototype' in newType && newType.prototype.render) {
          if (process.env.NODE_ENV === 'development') {
            throw new Error('Class component in render method is not supported.');
          }
        }
        newVNode._component = c = new Component(newProps);
        c.constructor = newType;
        c.render = doRender;
        c.props = newProps;
        isNew = c._dirty = true;
      }
      oldProps = c.props;
      c._vnode = newVNode;
      if (isNew) {
        // noop
      } else {
        if (newVNode._original === oldVNode._original) {
          newVNode._dom = oldVNode._dom;
          newVNode._children = oldVNode._children;
          newVNode._children.forEach(vnode => {
            if (vnode) vnode._parent = newVNode;
          });
          break outer;
        }
      }
      c.props = newProps;
      c._parentDom = parentDom;
      var renderHook = options._render;
      var count = 0;
      do {
        c._dirty = false;
        if (renderHook) renderHook(newVNode);
        tmp = c.render(c.props);

        // Handle state change in render
      } while (c._dirty && ++count < 25);
      var isTopLevelFragment = tmp != null && tmp.type === Fragment && tmp.key == null;
      var renderResult = isTopLevelFragment ? tmp.props.children : tmp;
      diffChildren(parentDom, Array.isArray(renderResult) ? renderResult : [renderResult], newVNode, oldVNode, isSvg, excessDomChildren, oldDom);
      c.base = newVNode._dom;
      if (clearProcessingException) {
        c._pendingError = c._processingException = null;
      }
    } else if (excessDomChildren == null && newVNode._original === oldVNode._original) {
      newVNode._children = oldVNode._children;
      newVNode._dom = oldVNode._dom;
    } else {
      newVNode._dom = diffElementNodes(oldVNode._dom, newVNode, oldVNode, isSvg, excessDomChildren);
    }
    if (tmp = options.diffed) tmp(newVNode);
  } catch (e) {
    newVNode._original = null;
    // if hydrating or creating initial tree, bailout preserves DOM:
    if (excessDomChildren != null) {
      newVNode._dom = oldDom;
      excessDomChildren[excessDomChildren.indexOf(oldDom)] = null;
      // ^ could possibly be simplified to:
      // excessDomChildren.length = 0;
    }

    options._catchError(e, newVNode, oldVNode);
  }
}

/**
 * Diff two virtual nodes representing DOM element
 * @param dom The DOM element representing
 * the virtual nodes being diffed
 * @param newVNode The new virtual node
 * @param oldVNode The old virtual node
 * @param {boolean} isSvg Whether or not this DOM node is an SVG node
 * @param excessDomChildren
 * @returns
 */
function diffElementNodes(dom, newVNode, oldVNode, isSvg, excessDomChildren) {
  var oldProps = oldVNode.props;
  var newProps = newVNode.props;
  var nodeType = newVNode.type;
  var i = 0;

  // Tracks entering and exiting SVG namespace when descending through the tree.
  if (nodeType === 'svg') isSvg = true;
  if (excessDomChildren != null) {
    for (; i < excessDomChildren.length; i++) {
      var child = excessDomChildren[i];

      // if newVNode matches an element in excessDomChildren or the `dom`
      // argument matches an element in excessDomChildren, remove it from
      // excessDomChildren so it isn't later removed in diffChildren
      if (child && 'setAttribute' in child === !!nodeType && (nodeType ? child.localName === nodeType : child.nodeType === 3)) {
        dom = child;
        excessDomChildren[i] = null;
        break;
      }
    }
  }
  if (dom == null) {
    if (nodeType === null) {
      // @ts-ignore createTextNode returns Text, we expect QuarkElement
      return document.createTextNode(newProps);
    }
    if (isSvg) {
      dom = document.createElementNS('http://www.w3.org/2000/svg',
      // @ts-ignore We know `newVNode.type` is a string
      nodeType);
    } else {
      dom = document.createElement(
      // @ts-ignore We know `newVNode.type` is a string
      nodeType, newProps.is && newProps);
    }

    // we created a new parent, so none of the previously attached children can be reused:
    excessDomChildren = null;
  }
  if (nodeType === null) {
    if (oldProps !== newProps) {
      dom.data = newProps;
    }
  } else {
    // If excessDomChildren was not null, repopulate it with the current element's children:
    excessDomChildren = excessDomChildren && slice.call(dom.childNodes);
    oldProps = oldVNode.props || EMPTY_OBJ;
    var oldHtml = oldProps.dangerouslySetInnerHTML;
    var newHtml = newProps.dangerouslySetInnerHTML;

    // TODO we should warn in debug mode when props don't match here.
    // But, if we are in a situation where we are using existing DOM (e.g. replaceNode)
    // we should read the existing DOM attributes to diff them
    if (excessDomChildren != null) {
      oldProps = {};
      for (i = 0; i < dom.attributes.length; i++) {
        oldProps[dom.attributes[i].name] = dom.attributes[i].value;
      }
    }
    if (newHtml || oldHtml) {
      // Avoid re-applying the same '__html' if it did not changed between re-render
      if (!newHtml || (!oldHtml || newHtml.__html != oldHtml.__html) && newHtml.__html !== dom.innerHTML) {
        dom.innerHTML = newHtml && newHtml.__html || '';
      }
    }
    diffProps(dom, newProps, oldProps, isSvg);

    // If the new vnode didn't have dangerouslySetInnerHTML, diff its children
    if (newHtml) {
      newVNode._children = [];
    } else {
      i = newVNode.props.children;
      diffChildren(dom, Array.isArray(i) ? i : [i], newVNode, oldVNode, isSvg && nodeType !== 'foreignObject', excessDomChildren, excessDomChildren ? excessDomChildren[0] : oldVNode._children && getDomSibling(oldVNode, 0));

      // Remove children that are not part of any vnode.
      if (excessDomChildren != null) {
        for (i = excessDomChildren.length; i--;) {
          if (excessDomChildren[i] != null) excessDomChildren[i].remove();
        }
      }
    }
    if ('value' in newProps && (i = newProps.value) !== undefined && (
    // For the <progress>-element the initial value is 0,
    // despite the attribute not being present. When the attribute
    // is missing the progress bar is treated as indeterminate.
    // To fix that we'll always update it when it is 0 for progress elements
    i !== dom.value || nodeType === 'progress' && !i ||
    // This is only for IE 11 to fix <select> value not being updated.
    // To avoid a stale select value we need to set the option.value
    // again, which triggers IE11 to re-evaluate the select value
    nodeType === 'option' && i !== oldProps.value)) {
      setProperty(dom, 'value', i, oldProps.value, false);
    }
    if ('checked' in newProps && (i = newProps.checked) !== undefined && i !== dom.checked) {
      setProperty(dom, 'checked', i, oldProps.checked, false);
    }
  }
  return dom;
}

/**
 * Invoke or update a ref, depending on whether it is a function or object ref.
 * @param {object|function} ref
 * @param {any} value
 * @param vnode
 */
function applyRef(ref, value, vnode) {
  try {
    if (typeof ref == 'function') ref(value);else ref.current = value;
  } catch (e) {
    options._catchError(e, vnode);
  }
}

/**
 * Unmount a virtual node from the tree and apply DOM changes
 * @param vnode The virtual node to unmount
 * @param parentVNode The parent of the VNode that
 * initiated the unmount
 * @param {boolean} [skipRemove] Flag that indicates that a parent node of the
 * current element is already detached from the DOM.
 */
function unmount(vnode, parentVNode, skipRemove) {
  var r;
  if (options.unmount) options.unmount(vnode);
  if (r = vnode.ref) {
    if (!r.current || r.current === vnode._dom) {
      applyRef(r, null, parentVNode);
    }
  }
  if ((r = vnode._component) != null) {
    r.base = r._parentDom = null;
    vnode._component = undefined;
  }
  if (r = vnode._children) {
    for (var i = 0; i < r.length; i++) {
      if (r[i]) {
        unmount(r[i], parentVNode, skipRemove || typeof vnode.type !== 'function');
      }
    }
  }
  if (!skipRemove && vnode._dom != null) {
    vnode._dom.remove();
  }

  // Must be set to `undefined` to properly clean up `_nextDom`
  // for which `null` is a valid value. See comment in `create-element.js`
  vnode._parent = vnode._dom = vnode._nextDom = undefined;
}

/** The `.render()` method for a PFC backing instance. */
function doRender(props) {
  return this.constructor(props);
}

/**
 * Render a quark virtual node into a DOM element
 * @param vnode The virtual node to render
 * @param root The root DOM element to render into
 */
function render(vnode, root) {
  if (options._root) options._root(vnode, root);

  // To be able to support calling `render()` multiple times on the same
  // DOM node, we need to obtain a reference to the previous tree. We do
  // this by assigning a new `_children` property to DOM nodes which points
  // to the last rendered tree. By default this property is not present, which
  // means that we are mounting a new tree for the first time.
  var oldVNode = root._children;
  vnode = root._children = createElement(Fragment, null, [vnode]);
  diff(root,
  // Determine the new vnode tree and store it on the DOM element on
  // our custom `_children` property.
  vnode, oldVNode || EMPTY_OBJ, root.ownerSVGElement !== undefined, oldVNode ? null : root.firstChild ? slice.call(root.childNodes) : null, oldVNode ? oldVNode._dom : root.firstChild);
}

class DblKeyMap {
  constructor() {
    _defineProperty(this, "map", new Map());
  }
  get(key, subKey) {
    var subMap = this.map.get(key);
    if (subMap) {
      if (!subKey) return subMap;
      return subMap.get(subKey);
    }
  }
  set(key, subKey, value) {
    var _subMap;
    var subMap = this.map.get(key);
    if (!subMap) {
      subMap = new Map();
      this.map.set(key, subMap);
    }
    (_subMap = subMap) === null || _subMap === void 0 ? void 0 : _subMap.set(subKey, value);
  }
  forEach(cb) {
    this.map.forEach((subMap, key) => {
      subMap.forEach((value, subKey) => {
        cb(value, key, subKey);
      });
    });
  }
  delete(key) {
    this.map.delete(key);
  }
  deleteAll() {
    this.map.forEach((_, key) => {
      this.map.delete(key);
    });
  }
}

class EventController {
  constructor() {
    _defineProperty(this, "eventMap", new DblKeyMap());
    _defineProperty(this, "bindListener", (el, eventName, eventHandler) => {
      if (!el || !eventName || !eventHandler) {
        return;
      }
      var list = this.eventMap.get(el, eventName);
      if (!list) {
        list = new Set();
        this.eventMap.set(el, eventName, list);
      }
      if (!list.has(eventHandler)) {
        list.add(eventHandler);
        el.addEventListener(eventName, eventHandler, true);
      }
    });
    _defineProperty(this, "removeAllListener", () => {
      this.eventMap.forEach((list, el, eventName) => {
        list.forEach(handler => {
          el.removeEventListener(eventName, handler, true);
        });
      });
      this.eventMap.deleteAll();
    });
  }
}

var version = "2.2.1";

function ownKeys$1(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); enumerableOnly && (symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; })), keys.push.apply(keys, symbols); } return keys; }
function _objectSpread$1(target) { for (var i = 1; i < arguments.length; i++) { var source = null != arguments[i] ? arguments[i] : {}; i % 2 ? ownKeys$1(Object(source), !0).forEach(function (key) { _defineProperty(target, key, source[key]); }) : Object.getOwnPropertyDescriptors ? Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)) : ownKeys$1(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } return target; }
/** dependency universal id */
var uid = 0;
/** current watcher that is collecting it's dependencies */
var currDepTarget;
/** all watchers waiting to run callback */
var watchers = [];
/** all waiting watchers' id */
var watcherIds = new Set();
/** are wathcers being flushed */
var flushing = false;
/** is flushing task already queued */
var flushWaiting = false;
/** flushing watcher index */
var flushIndex = 0;
/** is callbacks flushing task already queued */
var cbsFlushWaiting = false;
/** prepare micro task */
var queueMicroTask = callback => {
  if (typeof window.queueMicrotask === 'function') {
    window.queueMicrotask(callback);
  } else {
    Promise.resolve().then(callback);
  }
};
var cbs = [];
var flushCbs = () => {
  cbsFlushWaiting = false;
  var cbsToFlush = cbs.splice(0, cbs.length);
  for (var i = 0; i < cbsToFlush.length; i++) {
    cbsToFlush[i]();
  }
};
var nextTick = (cb, ctx) => {
  var _resolve;
  var _promise = !cb && new Promise(resolve => {
    _resolve = resolve;
  });
  cbs.push(() => {
    if (cb) {
      cb.call(ctx);
    } else {
      _resolve(ctx);
    }
  });
  if (!cbsFlushWaiting) {
    cbsFlushWaiting = true;
    queueMicroTask(flushCbs);
  }
  return _promise;
};
// const flushUpdatedQueue = (watchers: Watcher[]) => {
//   let i = watchers.length
//   while (i--) {
//     const watcher = watchers[i]
//     const {
//       render,
//       inst,
//     } = watcher
//     if (render && inst) {
//       inst.flushUpdatedQueue()
//     }
//   }
// }
/** flush watcher queue */
var flushWatcherQueue = () => {
  flushing = true;
  // make sure updates from parent to child, user watchers to render watchers.
  watchers.sort((a, b) => a.id - b.id);
  for (flushIndex = 0; flushIndex < watchers.length; flushIndex++) {
    var watcher = watchers[flushIndex];
    watcherIds.delete(watcher.id);
    watcher.run();
  }
  // flushUpdatedQueue(watchers)
  // reset
  watchers.length = 0;
  watcherIds.clear();
  flushing = false;
  flushWaiting = false;
};
/** add watcher to queue */
var queueWatcher = watcher => {
  var {
    id
  } = watcher;
  if (watcherIds.has(id)) {
    return;
  }
  watcherIds.add(id);
  if (!flushing) {
    watchers.push(watcher);
  } else {
    var placeIndex = watchers.findIndex(target => target.id > id);
    if (placeIndex === -1) {
      placeIndex = watchers.length;
    } else if (placeIndex > flushIndex) {
      placeIndex = flushIndex + 1;
    }
    watchers.splice(placeIndex, 0, watcher);
  }
  if (!flushWaiting) {
    flushWaiting = true;
    nextTick(flushWatcherQueue);
  }
};
class Watcher {
  constructor(inst, getterOrExpr, options) {
    /** watcher's id */
    _defineProperty(this, "id", void 0);
    /**
     * watcher itself can be dependency of others'
     *
     * so you can treat it as watcher itself (for maintaining dependencies)
     */
    _defineProperty(this, "dep", new Dep());
    /** this watcher's dependencies */
    _defineProperty(this, "deps", new Map());
    _defineProperty(this, "oldDeps", new Map());
    /** watcher's getter function */
    _defineProperty(this, "getter", void 0);
    /** quark element's instance */
    _defineProperty(this, "inst", void 0);
    /** computed value */
    _defineProperty(this, "value", undefined);
    /** is computed value dirty */
    _defineProperty(this, "dirty", true);
    /** is computed watcher */
    _defineProperty(this, "computed", void 0);
    /** watcher's callback */
    _defineProperty(this, "cb", void 0);
    /** is render watcher */
    _defineProperty(this, "render", void 0);
    this.id = ++uid;
    this.inst = inst;
    var defaultOptions = {
      cb: noop,
      render: false,
      computed: false,
      immediate: false
    };
    var {
      computed,
      render,
      cb,
      immediate
    } = _objectSpread$1(_objectSpread$1({}, defaultOptions), options);
    this.computed = computed;
    this.render = render;
    if (typeof getterOrExpr === 'function') {
      this.getter = getterOrExpr;
    } else {
      this.getter = () => {
        return getterOrExpr.split('.').reduce((acc, pathSeg) => acc === null || acc === void 0 ? void 0 : acc[pathSeg], inst);
      };
    }
    this.cb = cb;
    if (!computed) {
      if (!immediate) {
        this.value = this.get();
      } else {
        this.run();
      }
    }
  }
  /** get latest computed value of watcher */
  get() {
    if (this.dirty) {
      this.value = this.compute();
      this.dirty = false;
    }
    return this.value;
  }
  /** invoke getter, recompute value, manage dependencies */
  compute() {
    setCurrDepTarget(this);
    // reset deps collection
    this.oldDeps = this.deps;
    this.deps = new Map();
    // invoke getter and collect new dependencies
    var value = this.getter.call(this.inst);
    popCurrDepTarget();
    this.cleanDeps();
    return value;
  }
  /** clean invalidated dependencies */
  cleanDeps() {
    this.oldDeps.forEach((dep, id) => {
      if (!this.deps.has(id)) {
        dep.unwatch(this);
      }
    });
  }
  /** add dependency to this watcher's dependency list */
  addDep(dep) {
    var {
      id
    } = dep;
    if (!this.deps.has(id)) {
      this.deps.set(id, dep);
      dep.watch(this);
    }
  }
  /** mark itself as a dependency of current watcher */
  depend() {
    if (currDepTarget) {
      currDepTarget.addDep(this.dep);
    }
  }
  /** invoke callback after successfully updated */
  updateAndInvoke(cb) {
    var value = this.compute();
    var oldValue = this.value;
    if (Object.is(value, oldValue)) {
      return;
    }
    this.value = value;
    this.dirty = false;
    cb.call(this.inst, value, oldValue);
  }
  /** receive update notification from it's dependency */
  update() {
    if (this.computed) {
      if (!this.dep.watchers.size) {
        // if no watcher is watching this computed value, simply mark it dirty
        // its watchers' count may change during app's execution
        this.dirty = true;
      } else {
        // * notify watchers of this watcher（for computed）
        this.updateAndInvoke(() => {
          this.dep.notify();
        });
      }
    } else {
      queueWatcher(this);
    }
  }
  /** run callback of watcher */
  run() {
    this.updateAndInvoke(this.cb);
  }
}
class Dep {
  constructor() {
    /** dependecy id */
    _defineProperty(this, "id", void 0);
    /** watchers that are watching this dependecy, listening to its changes */
    _defineProperty(this, "watchers", void 0);
    this.id = uid++;
    this.watchers = new Set();
  }
  /** a watcher subscribe to this dependency */
  watch(watcher) {
    this.watchers.add(watcher);
  }
  /** stop watching this dependency */
  unwatch(watcher) {
    this.watchers.delete(watcher);
  }
  /** notify update to its watchers */
  notify() {
    this.watchers.forEach(watcher => {
      watcher.update();
    });
  }
  /**
   * mark itself as a dependency of current watcher.
   * in other words, current watcher depends on it.
  */
  depend() {
    if (currDepTarget) {
      currDepTarget.addDep(this);
    }
  }
}
var depTargetStack = [];
/** setter for {@link currDepTarget} */
var setCurrDepTarget = target => {
  depTargetStack.push(target);
  currDepTarget = target;
};
var popCurrDepTarget = () => {
  depTargetStack.pop();
  currDepTarget = depTargetStack[depTargetStack.length - 1];
};

var _excluded = ["path"];
function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); enumerableOnly && (symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; })), keys.push.apply(keys, symbols); } return keys; }
function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = null != arguments[i] ? arguments[i] : {}; i % 2 ? ownKeys(Object(source), !0).forEach(function (key) { _defineProperty(target, key, source[key]); }) : Object.getOwnPropertyDescriptors ? Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)) : ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } return target; }
function createRef() {
  return {
    current: null
  };
}
if (process.env.NODE_ENV === 'development') {
  console.info("%cquarkc@".concat(version), 'color: white;background:#9f57f8;font-weight:bold;font-size:10px;padding:2px 6px;border-radius: 5px', 'Running in dev mode.');
}
/**
 * Check if val is null or undefined.
 *
 * 判断是否为空值，
 */
var isEmpty = val => val == null;
var defaultPropertyDeclaration = {
  observed: true,
  type: String,
  converter: (value, type) => {
    switch (type) {
      case Number:
        return Number(value);
      case Boolean:
        return value !== null && value !== 'false';
      // noop
    }

    return value;
  }
};
var property = function () {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  return (target, propName) => {
    return target.constructor.createProperty(propName, options);
  };
};
var internalProp = () => {
  return (target, propName) => {
    return target.constructor.createProperty(propName, {
      internal: true
    });
  };
};
var state = () => {
  return (target, propName) => {
    return target.constructor.createState(propName);
  };
};
var computed = () => {
  return (target, propName, descriptor) => {
    return target.constructor.computed(propName, descriptor);
  };
};
var watch = (path, options) => {
  return (target, propName, descriptor) => {
    return target.constructor.watch(propName, descriptor, path, options);
  };
};
var StateDescriptors = new DblKeyMap();
/** all declared props' definitions, with attribute name as sub key */
var PropDefs = new DblKeyMap();
/** quark element instance's props' map, with attribute name as sub key */
var Props = new DblKeyMap();
var ComputedDescriptors = new DblKeyMap();
var UserWatchers = new DblKeyMap();
function getWrapperClass(target, style) {
  return class QuarkElementWrapper extends target {
    static get _observedAttrs() {
      var _PropDefs$get;
      return [...(((_PropDefs$get = PropDefs.get(target)) === null || _PropDefs$get === void 0 ? void 0 : _PropDefs$get.entries()) || [])].filter(_ref => {
        var [_, {
          options
        }] = _ref;
        return !!options.observed;
      });
    }
    /** ! HTML SPEC field: specify the attribute names to watch for changes */
    static get observedAttributes() {
      return this._observedAttrs.map(_ref2 => {
        var [attrName] = _ref2;
        return attrName;
      });
    }
    static _isInternalProp(propName) {
      var _PropDefs$get2;
      var def = (_PropDefs$get2 = PropDefs.get(target)) === null || _PropDefs$get2 === void 0 ? void 0 : _PropDefs$get2.get(propName);
      return def ? !!def.options.internal : false;
    }
    static _isBoolProp(attrName) {
      var _PropDefs$get3;
      var def = (_PropDefs$get3 = PropDefs.get(target)) === null || _PropDefs$get3 === void 0 ? void 0 : _PropDefs$get3.get(attrName);
      if (!def) {
        return false;
      }
      return def.options.type === Boolean;
    }
    constructor() {
      var _this;
      super();
      _this = this;
      var shadowRoot = this.attachShadow({
        mode: "open"
      });
      if (shadowRoot) {
        // Create Css
        if (typeof CSSStyleSheet === "function" && shadowRoot.adoptedStyleSheets) {
          // Use constructed style first
          var sheet = new CSSStyleSheet();
          sheet.replaceSync(style);
          shadowRoot.adoptedStyleSheets = [sheet];
        } else {
          // Fallback
          var styleEl = document.createElement("style");
          styleEl.innerHTML = style;
          shadowRoot.append(styleEl);
        }
      }
      // * 获取包装类实例的父类（即继承了QuarkElement的类——用户书写的组件类）
      // * get parent class (user-defined component class that extends QuarkElement) of wrapper class
      var UserComp = Object.getPrototypeOf(this.constructor);
      var stateDescriptors = StateDescriptors.get(UserComp);
      if (stateDescriptors !== null && stateDescriptors !== void 0 && stateDescriptors.size) {
        stateDescriptors.forEach((descriptorCreator, propName) => {
          Object.defineProperty(this, propName, descriptorCreator(this[propName]));
        });
      }
      /**
       * 重写类的属性描述符，并重写属性初始值。
       * 注：由于子类的属性初始化晚于当前基类的构造函数，同名属性会导致属性描述符被覆盖，所以必须放在基类构造函数之后执行
       */
      var propDefs = PropDefs.get(UserComp);
      if (propDefs !== null && propDefs !== void 0 && propDefs.size) {
        propDefs.forEach((def, attrName) => {
          var {
            options: {
              type,
              converter,
              internal
            },
            propName
          } = def;
          var defaultValue = this[propName];
          if (internal) {
            Object.defineProperty(this, propName, UserComp.getStateDescriptor(propName)(defaultValue));
            return;
          }
          var isBoolProp = type === Boolean;
          var hasConverter = isFunction(converter);
          /** convert attribute's value to its decorated counterpart, that is, property's value */
          var convertAttrValue = value => {
            // * For boolean properties, ignore the defaultValue specified.
            // * We should always respect to whether the boolean attribute is set.
            if (!isBoolProp && isEmpty(value)) {
              return defaultValue;
            }
            if (!hasConverter) {
              return value;
            }
            return converter(value, type);
          };
          var dep = new Dep();
          Props.set(this, attrName, {
            propName,
            dep,
            converter: convertAttrValue
          });
          // make attribute reactive
          Object.defineProperty(this, propName, {
            get() {
              dep.depend();
              return convertAttrValue(this.getAttribute(attrName));
            },
            set(val) {
              updateDomAttr(this, attrName, val);
            },
            configurable: true,
            enumerable: true
          });
        });
      }
      var computedDescriptors = ComputedDescriptors.get(UserComp);
      if (computedDescriptors !== null && computedDescriptors !== void 0 && computedDescriptors.size) {
        computedDescriptors.forEach((descriptorCreator, propKey) => {
          Object.defineProperty(this, propKey, descriptorCreator());
        });
      }
      var watchers = UserWatchers.get(UserComp);
      if (watchers !== null && watchers !== void 0 && watchers.size) {
        watchers.forEach((_ref3, propName) => {
          var {
              path
            } = _ref3,
            options = _objectWithoutProperties(_ref3, _excluded);
          new Watcher(this, path, _objectSpread(_objectSpread({}, options), {}, {
            cb: function () {
              return _this[propName](...arguments);
            }
          }));
        });
      }
    }
  };
}
function customElement(params) {
  var {
    tag,
    style = ''
  } = typeof params === 'string' ? {
    tag: params
  } : params;
  return target => {
    if (!customElements.get(tag)) {
      customElements.define(tag, getWrapperClass(target, style));
    }
  };
}
class QuarkElement extends HTMLElement {
  constructor() {
    super(...arguments);
    _defineProperty(this, "_updatedQueue", []);
    /** 组件是否已挂载 */
    _defineProperty(this, "_mounted", false);
    _defineProperty(this, "_renderWatcher", undefined);
    _defineProperty(this, "eventController", new EventController());
    _defineProperty(this, "_controllers", void 0);
    _defineProperty(this, "rootPatch", newRootVNode => {
      if (this.shadowRoot) {
        render(newRootVNode, this.shadowRoot);
      }
    });
    _defineProperty(this, "$on", (eventName, eventHandler, el) => {
      return this.eventController.bindListener(el || this, eventName, eventHandler);
    });
    /** log old 'false' attribute value before resetting and removing it */
    _defineProperty(this, "_oldVals", new Map());
  }
  queueUpdated(cb) {
    this._updatedQueue.push(cb);
  }
  hasOwnLifeCycleMethod(methodName) {
    return Object.getPrototypeOf(this.constructor // base class
    ) // wrapper class
    .prototype // user-defined class
    .hasOwnProperty(methodName);
  }
  hasDidUpdateCb() {
    return this.hasOwnLifeCycleMethod('componentDidUpdate');
  }
  /** handler for processing tasks after render */
  postRender() {
    var mounted = this._mounted;
    if (!mounted) {
      this._mounted = true;
      var hasPendingUpdate = !!this._updatedQueue.length;
      if (this.hasOwnLifeCycleMethod('componentDidMount')) {
        if (hasPendingUpdate) {
          // * when componentDidMount and componentDidUpate are both defined
          // * componentDidUpate should be ignored at mount phase
          this._updatedQueue = [];
          this.componentDidMount();
          return;
        }
        this.componentDidMount();
      } else {
        // * for historic reasons, componentDidUpate was used for both mounting and updating
        // * so we should also call it at mount phase
        if (process.env.NODE_ENV === 'development') {
          if (hasPendingUpdate) {
            console.warn('by design, componentDidUpdate should not be invoked at mount phase, use componentDidMount for initialization logic instead.');
          }
        }
      }
    }
    this._updatedQueue.forEach(cb => cb());
    this._updatedQueue = [];
    if (mounted) {
      if (this.hasOwnLifeCycleMethod('componentUpdated')) {
        this.componentUpdated();
      }
    }
  }
  // 内部属性装饰器
  static getStateDescriptor(propName) {
    return defaultValue => {
      var value = defaultValue;
      var dep;
      var getDep = () => dep || (dep = new Dep());
      return {
        get() {
          getDep().depend();
          return value;
        },
        set(newValue) {
          var resolvedOldVal = value;
          if (Object.is(resolvedOldVal, newValue)) {
            return;
          }
          if (this.shouldPreventUpdate(propName, resolvedOldVal, newValue)) {
            return;
          }
          value = newValue;
          getDep().notify();
          if (this.hasDidUpdateCb()) {
            this.queueUpdated(() => {
              this.componentDidUpdate(propName, resolvedOldVal, newValue);
            });
          }
        },
        configurable: true,
        enumerable: true
      };
    };
  }
  static createProperty(propName, options) {
    var attrName = options.attribute || propName;
    PropDefs.set(this, attrName, {
      options: _objectSpread(_objectSpread(_objectSpread({}, defaultPropertyDeclaration), options), options.internal ? {
        observed: false
      } : null),
      propName
    });
  }
  static createState(propName) {
    StateDescriptors.set(this, propName, this.getStateDescriptor(propName));
  }
  static computed(propName, descriptor) {
    if (descriptor.get) {
      ComputedDescriptors.set(this, propName, () => {
        var watcher;
        return {
          configurable: true,
          enumerable: true,
          get() {
            if (!watcher) {
              watcher = new Watcher(this, descriptor.get, {
                computed: true
              });
            }
            watcher.dep.depend();
            return watcher.get();
          }
        };
      });
    }
  }
  static watch(propName, descriptor, path, options) {
    if (isFunction(descriptor.value)) {
      UserWatchers.set(this, propName, _objectSpread(_objectSpread({}, options), {}, {
        path
      }));
    }
  }
  _render() {
    var newRootVNode = this.render();
    this.rootPatch(newRootVNode);
  }
  /** sync property with its attribute counterpart */
  _updateObservedProps() {
    this.constructor._observedAttrs.forEach(_ref4 => {
      var [_, {
        propName
      }] = _ref4;
      this[propName] = this[propName];
    });
  }
  /**
   * Registers a `ReactiveController` to participate in the element's reactive
   * update cycle. The element automatically calls into any registered
   * controllers during its lifecycle callbacks.
   *
   * If the element is connected when `addController()` is called, the
   * controller's `hostConnected()` callback will be immediately called.
   * @category controllers
   */
  addController(controller) {
    var _this$_controllers;
    ((_this$_controllers = this._controllers) !== null && _this$_controllers !== void 0 ? _this$_controllers : this._controllers = new Set()).add(controller);
    if (this.shadowRoot && this.isConnected) {
      var _controller$hostConne;
      (_controller$hostConne = controller.hostConnected) === null || _controller$hostConne === void 0 ? void 0 : _controller$hostConne.call(controller);
    }
  }
  /**
   * Removes a `ReactiveController` from the element.
   * @category controllers
   */
  removeController(controller) {
    var _this$_controllers2;
    (_this$_controllers2 = this._controllers) === null || _this$_controllers2 === void 0 ? void 0 : _this$_controllers2.delete(controller);
  }
  // Reserve, may expand in the future
  requestUpdate() {
    this.getOrInitRenderWatcher().update();
  }
  // Reserve, may expand in the future
  update() {
    this.getOrInitRenderWatcher().update();
  }
  $emit(eventName, customEventInit) {
    return this.dispatchEvent(new CustomEvent(eventName, Object.assign({
      bubbles: true
    }, customEventInit)));
  }
  $nextTick(cb) {
    return nextTick(cb, this);
  }
  /**
   * 此时组件 dom 已插入到页面中，等同于 connectedCallback() { super.connectedCallback(); }
   */
  componentDidMount() {}
  /**
   * disconnectedCallback 触发时、dom 移除前执行，等同于 disconnectedCallback() { super.disconnectedCallback(); }
   */
  componentWillUnmount() {}
  /**
   * @deprecated
   * since we have embraced more precisely controlled render scheduler mechanism,
   * there's no need to use shouldComponentUpdate any more,
   * and will be removed in next major version.
   *
   * 控制当前属性变化是否导致组件渲染
   * @param propName 属性名
   * @param oldValue 属性旧值
   * @param newValue 属性新值
   * @returns boolean
   */
  shouldComponentUpdate(propName, oldValue, newValue) {
    return oldValue !== newValue;
  }
  shouldPreventUpdate(propName, oldValue, newValue) {
    if (this.hasOwnLifeCycleMethod('shouldComponentUpdate')) {
      return !this.shouldComponentUpdate(propName, oldValue, newValue);
    }
    return false;
  }
  /** @deprecated use \@watch directive instead for same purposes */
  componentDidUpdate(propName, oldValue, newValue) {}
  /** called when all props and states updated */
  componentUpdated() {}
  /**
   * 组件的 render 方法，
   * 自动执行 this.shadowRoot.innerHTML = this.render()
   * @returns VNode
   */
  render() {
    return "";
  }
  getOrInitRenderWatcher() {
    if (!this._renderWatcher) {
      this._renderWatcher = new Watcher(this, () => {
        var _this$_controllers3;
        this._render();
        var renderCbType = this._mounted ? 'hostUpdated' : 'hostMounted';
        (_this$_controllers3 = this._controllers) === null || _this$_controllers3 === void 0 ? void 0 : _this$_controllers3.forEach(c => {
          var _c$renderCbType;
          return (_c$renderCbType = c[renderCbType]) === null || _c$renderCbType === void 0 ? void 0 : _c$renderCbType.call(c);
        });
        this.postRender();
      }, {
        render: true
      });
    }
    return this._renderWatcher;
  }
  connectedCallback() {
    var _this$_controllers4;
    this._updateObservedProps();
    (_this$_controllers4 = this._controllers) === null || _this$_controllers4 === void 0 ? void 0 : _this$_controllers4.forEach(c => {
      var _c$hostConnected;
      return (_c$hostConnected = c.hostConnected) === null || _c$hostConnected === void 0 ? void 0 : _c$hostConnected.call(c);
    });
    this.getOrInitRenderWatcher();
  }
  attributeChangedCallback(attrName, oldVal, newVal) {
    var _Props$get;
    var prop = (_Props$get = Props.get(this)) === null || _Props$get === void 0 ? void 0 : _Props$get.get(attrName);
    if (!prop) {
      return;
    }
    var {
      propName
    } = prop;
    // React specific patch, for more detailed explanation: https://github.com/facebook/react/issues/9230
    // But also works for unintentionally set attribute value to string 'false', it's not recommended in HTML spec.
    // For custom elements, react will pass down the boolean value as-is, that is,
    // the attribute value will be string 'true' or 'false', they all will be treated as true by HTML spec.
    // We should remove the attribute when its value is 'false' to prevent ambiguity and line up with HTML spec.
    // One more thing, in CSS boolean attribute with value 'false' will match the attribute selector [attr].
    // 
    // 对于自定义元素，React会直接将布尔值属性传递下去
    // 这时候这里的value会是字符串'true'或'false'，对于'false'我们需要手动将该属性从自定义元素上移除
    // 以避免CSS选择器将[attr="false"]视为等同于[attr]
    if (newVal !== oldVal) {
      if (this.constructor._isBoolProp(attrName)) {
        if (newVal === 'false' && attrName.indexOf('aria-') === -1) {
          if (this.hasDidUpdateCb()) {
            this._oldVals.set(propName, oldVal);
          }
          // manually remove the attribute
          this[propName] = false;
          return;
        }
      }
    }
    var newValue = this[propName];
    var resolvedOldVal = oldVal;
    var oldValReset = this._oldVals.get(propName);
    if (oldValReset) {
      this._oldVals.delete(propName);
      resolvedOldVal = oldValReset;
    }
    resolvedOldVal = prop.converter(resolvedOldVal);
    if (this.shouldPreventUpdate(propName, resolvedOldVal, newValue)) {
      return;
    }
    // notify changes to this prop's watchers
    prop.dep.notify();
    if (this.hasDidUpdateCb()) {
      this.queueUpdated(() => {
        this.componentDidUpdate(propName, resolvedOldVal, newValue);
      });
    }
  }
  disconnectedCallback() {
    var _this$_controllers5;
    if (this.hasOwnLifeCycleMethod('componentWillUnmount')) {
      this.componentWillUnmount();
    }
    Props.delete(this);
    this.eventController.removeAllListener();
    (_this$_controllers5 = this._controllers) === null || _this$_controllers5 === void 0 ? void 0 : _this$_controllers5.forEach(c => {
      var _c$hostDisconnected;
      return (_c$hostDisconnected = c.hostDisconnected) === null || _c$hostDisconnected === void 0 ? void 0 : _c$hostDisconnected.call(c);
    });
    this.rootPatch(null);
    this._mounted = false;
  }
}
_defineProperty(QuarkElement, "h", createElement);
_defineProperty(QuarkElement, "Fragment", Fragment);

export { Fragment, QuarkElement, computed, createRef, customElement, internalProp, property, state, watch };
